#include <linux/config.h>

#ifdef CONFIG_USB_DEBUG
#define DEBUG
#endif

#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/completion.h>
#include <linux/usb.h>
#include <linux/diskdump.h>

#include "usb.h"
#include "hcd.h"


void usb_hcd_poll(struct usb_hcd *hcd)
{
	unsigned long flags;

	local_irq_save(flags);
	usb_hcd_irq(0, hcd, NULL);
	local_irq_restore(flags);
}
EXPORT_SYMBOL (usb_hcd_poll);

void usb_dump_wait_for_completion(struct completion *x, struct urb *urb)
{
	unsigned long flags;

	might_sleep();
	spin_lock_irqsave(&x->wait.lock, flags);
	if (!x->done) {
		struct usb_hcd *hcd = urb->dev->bus->hcpriv;

		DECLARE_WAITQUEUE(wait, current);

		wait.flags |= WQ_FLAG_EXCLUSIVE;
		__add_wait_queue_tail(&x->wait, &wait);
		do {
			__set_current_state(TASK_UNINTERRUPTIBLE);
			spin_unlock_irqrestore(&x->wait.lock, flags);
			usb_hcd_poll(hcd);
			udelay(100);
			diskdump_update();
			spin_lock_irqsave(&x->wait.lock, flags);
		} while (!x->done);
		__remove_wait_queue(&x->wait, &wait);
	}
	x->done--;
	spin_unlock_irqrestore(&x->wait.lock, flags);
}
EXPORT_SYMBOL (usb_dump_wait_for_completion);

