/*
 * Copyright (c) 2004,2005 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Further, this software is distributed without any warranty that it is
 * free of the rightful claim of any third person regarding infringement
 * or the like.  Any license provided herein, whether implied or
 * otherwise, applies only to this software file.  Patent licenses, if
 * any, provided herein do not apply to combinations of this program with
 * other software, or any other product whatsoever.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
 * Mountain View, CA  94043, or:
 *
 * http://www.sgi.com
 *
 * For further information regarding this notice, see:
 *
 * http://oss.sgi.com/projects/GenInfo/SGIGPLNoticeExplan/
 */

#include <linux/config.h>
#include <linux/compat.h>
#include <linux/init.h>
#include <linux/ioctl.h>
#include <linux/ioctl32.h>
#include <linux/syscalls.h>
#include <linux/types.h>
#include <asm/uaccess.h>

#include "xfs_types.h"
#include "xfs_fs.h"
#include "xfs_dfrag.h"

#define  _NATIVE_IOC(cmd, type) \
	   _IOC(_IOC_DIR(cmd), _IOC_TYPE(cmd), _IOC_NR(cmd), sizeof(type))

#if defined(CONFIG_IA64) || defined(CONFIG_X86_64)
#define BROKEN_X86_ALIGNMENT

/* on ia32 l_start is on a 32-bit boundary */
typedef struct xfs_flock64_32 {
	__s16		l_type;
	__s16		l_whence;
	__s64		l_start	__attribute__((packed));
				/* len == 0 means until end of file */
	__s64		l_len	__attribute__((packed));
	__s32		l_sysid;
	__u32		l_pid;
	__s32		l_pad[4];	/* reserve area	*/
} xfs_flock64_32_t;

#define XFS_IOC_ALLOCSP_32	_IOW ('X', 10, struct xfs_flock64_32)
#define XFS_IOC_FREESP_32	_IOW ('X', 11, struct xfs_flock64_32)
#define XFS_IOC_ALLOCSP64_32	_IOW ('X', 36, struct xfs_flock64_32)
#define XFS_IOC_FREESP64_32	_IOW ('X', 37, struct xfs_flock64_32)
#define XFS_IOC_RESVSP_32	_IOW ('X', 40, struct xfs_flock64_32)
#define XFS_IOC_UNRESVSP_32	_IOW ('X', 41, struct xfs_flock64_32)
#define XFS_IOC_RESVSP64_32	_IOW ('X', 42, struct xfs_flock64_32)
#define XFS_IOC_UNRESVSP64_32	_IOW ('X', 43, struct xfs_flock64_32)

/* copy xfs_flock64 into aligned version */
static int
xfs_ioctl32_flock(
	unsigned int		fd,
	unsigned int		cmd,
	unsigned long		arg,
	struct file *		file)
{
	xfs_flock64_32_t	__user *p32 = (void __user *)arg;
	xfs_flock64_t		__user *p = compat_alloc_user_space(sizeof(*p));

	if (copy_in_user(&p->l_type, &p32->l_type, sizeof(s16)) ||
	    copy_in_user(&p->l_whence, &p32->l_whence, sizeof(s16)) ||
	    copy_in_user(&p->l_start, &p32->l_start, sizeof(s64)) ||
	    copy_in_user(&p->l_len, &p32->l_len, sizeof(s64)) ||
	    copy_in_user(&p->l_sysid, &p32->l_sysid, sizeof(s32)) ||
	    copy_in_user(&p->l_pid, &p32->l_pid, sizeof(u32)) ||
	    copy_in_user(&p->l_pad, &p32->l_pad, 4*sizeof(s32)))
		return -EFAULT;

	return sys_ioctl(fd, _NATIVE_IOC(cmd, struct xfs_flock64),
			(unsigned long)p);
}

#else

/* no alignment problems for flock_t */
#define xfs_ioctl32_flock NULL

#define XFS_IOC_ALLOCSP_32	XFS_IOC_ALLOCSP
#define XFS_IOC_FREESP_32	XFS_IOC_FREESP
#define XFS_IOC_ALLOCSP64_32	XFS_IOC_ALLOCSP64
#define XFS_IOC_FREESP64_32	XFS_IOC_FREESP64
#define XFS_IOC_RESVSP_32	XFS_IOC_RESVSP
#define XFS_IOC_UNRESVSP_32	XFS_IOC_UNRESVSP
#define XFS_IOC_RESVSP64_32	XFS_IOC_RESVSP64
#define XFS_IOC_UNRESVSP64_32	XFS_IOC_UNRESVSP64

typedef struct xfs_fsop_bulkreq32 {
	compat_uptr_t	lastip;		/* last inode # pointer		*/
	__s32		icount;		/* count of entries in buffer	*/
	compat_uptr_t	ubuffer;	/* user buffer for inode desc.	*/
	__s32		ocount;		/* output count pointer		*/
} xfs_fsop_bulkreq32_t;

static int
xfs_ioctl32_bulkstat(
	unsigned int		fd,
	unsigned int		cmd,
	unsigned long		arg,
	struct file *		file)
{
	xfs_fsop_bulkreq32_t	__user *p32 = (void __user *)arg;
	xfs_fsop_bulkreq_t	__user *p = compat_alloc_user_space(sizeof(*p));
	u32			addr;

	if (get_user(addr, &p32->lastip) ||
	    put_user(compat_ptr(addr), &p->lastip) ||
	    copy_in_user(&p->icount, &p32->icount, sizeof(s32)) ||
	    get_user(addr, &p32->ubuffer) ||
	    put_user(compat_ptr(addr), &p->ubuffer) ||
	    get_user(addr, &p32->ocount) ||
	    put_user(compat_ptr(addr), &p->ocount))
		return -EFAULT;

	return sys_ioctl(fd, cmd, (unsigned long)p);
}
#endif

struct ioctl_trans xfs_ioctl32_trans[] = {
	{ XFS_IOC_DIOINFO, },
	{ XFS_IOC_FSGEOMETRY_V1, },
	{ XFS_IOC_FSGEOMETRY, },
	{ XFS_IOC_GETVERSION, },
	{ XFS_IOC_GETXFLAGS, },
	{ XFS_IOC_SETXFLAGS, },
	{ XFS_IOC_FSGETXATTR, },
	{ XFS_IOC_FSSETXATTR, },
	{ XFS_IOC_FSGETXATTRA, },
	{ XFS_IOC_FSSETDM, },
	{ XFS_IOC_GETBMAP, },
	{ XFS_IOC_GETBMAPA, },
	{ XFS_IOC_GETBMAPX, },
/* not handled
	{ XFS_IOC_FD_TO_HANDLE, },
	{ XFS_IOC_PATH_TO_HANDLE, },
	{ XFS_IOC_PATH_TO_HANDLE, },
	{ XFS_IOC_PATH_TO_FSHANDLE, },
	{ XFS_IOC_OPEN_BY_HANDLE, },
	{ XFS_IOC_FSSETDM_BY_HANDLE, },
	{ XFS_IOC_READLINK_BY_HANDLE, },
	{ XFS_IOC_ATTRLIST_BY_HANDLE, },
	{ XFS_IOC_ATTRMULTI_BY_HANDLE, },
*/
	{ XFS_IOC_FSCOUNTS, NULL, },
	{ XFS_IOC_SET_RESBLKS, NULL, },
	{ XFS_IOC_GET_RESBLKS, NULL, },
	{ XFS_IOC_FSGROWFSDATA, NULL, },
	{ XFS_IOC_FSGROWFSLOG, NULL, },
	{ XFS_IOC_FSGROWFSRT, NULL, },
	{ XFS_IOC_FREEZE, NULL, },
	{ XFS_IOC_THAW, NULL, },
	{ XFS_IOC_GOINGDOWN, NULL, },
	{ XFS_IOC_ERROR_INJECTION, NULL, },
	{ XFS_IOC_ERROR_CLEARALL, NULL, },
	/* xfs_flock_t has wrong u32 vs u64 alignment */
	{ XFS_IOC_ALLOCSP_32, xfs_ioctl32_flock },
	{ XFS_IOC_FREESP_32, xfs_ioctl32_flock },
	{ XFS_IOC_RESVSP_32, xfs_ioctl32_flock },
	{ XFS_IOC_UNRESVSP_32, xfs_ioctl32_flock },
	{ XFS_IOC_ALLOCSP64_32, xfs_ioctl32_flock },
	{ XFS_IOC_FREESP64_32, xfs_ioctl32_flock },
	{ XFS_IOC_RESVSP64_32, xfs_ioctl32_flock },
	{ XFS_IOC_UNRESVSP64_32, xfs_ioctl32_flock },
#ifndef BROKEN_X86_ALIGNMENT
	/* xfs_bstat_t has wrong u32 vs u64 alignment */
	{ XFS_IOC_SWAPEXT, },
	{ XFS_IOC_FSBULKSTAT_SINGLE, xfs_ioctl32_bulkstat },
	{ XFS_IOC_FSBULKSTAT, xfs_ioctl32_bulkstat},
	{ XFS_IOC_FSINUMBERS, xfs_ioctl32_bulkstat},
#endif
	{ 0, },
};

int __init
xfs_ioctl32_init(void)
{
	int error, i;

	for (i = 0; xfs_ioctl32_trans[i].cmd != 0; i++) {
		error = register_ioctl32_conversion(xfs_ioctl32_trans[i].cmd,
				xfs_ioctl32_trans[i].handler);
		if (error)
			goto fail;
	}

	return 0;

 fail:
	while (--i)
		unregister_ioctl32_conversion(xfs_ioctl32_trans[i].cmd);
	return error;
}

void
xfs_ioctl32_exit(void)
{
	int i;

	for (i = 0; xfs_ioctl32_trans[i].cmd != 0; i++)
		unregister_ioctl32_conversion(xfs_ioctl32_trans[i].cmd);
}
